﻿// ECMAScript 5 strict mode
"use strict";

assert2(cr, "cr namespace not created");
assert2(cr.plugins_, "cr.plugins_ not created");

/// Plugin class
cr.plugins_.HtmEl = function(runtime)
{
	this.runtime = runtime;
};

(function ()
{
	//:	const
	var RIN_SEPARATOR = /\s+/g;

	//:	vars
	var last_canvas_offset = null;
	var last_checked_tick = -1;

	var pluginProto = cr.plugins_.HtmEl.prototype;
		

	/// Object type class
	pluginProto.Type = function(plugin)
	{
		this.plugin = plugin;
		this.runtime = plugin.runtime;
	};

	var typeProto = pluginProto.Type.prototype;

	/// called on startup for each object type
	typeProto.onCreate = function()
	{
	};

	/// Instance class
	pluginProto.Instance = function(type)
	{
		this.type = type;
		this.runtime = type.runtime;
	};
	
	var instanceProto = pluginProto.Instance.prototype;




	/// called whenever an instance is created
	instanceProto.onCreate = function()
	{
		//	Not supported in directCanvas
		//	TOD?	this.runtime.isDomFree === no Dom exist?	
		if (this.runtime.isDomFree)
		{
			cr.logexport("[Construct 2] HtmEl plugin not supported on this platform - the object will not be created");
			return;
		}
		
		//:	possibly c2 does not care about name confilct? TODM?
		//	so, we create names whichever we want ... doctor will take care after accident ...
		this.ret = null;

		this.elem		= document.createElement( this.properties[0] );
		this.jqelem		= jQuery( this.elem );
			
		this.labelText = null;
		
		if( this.properties[6] ) this.elem.id = this.properties[6];
		if( this.properties[1] ) this.elem.innerHTML = this.properties[1];
		if( this.properties[7] ) this.elem.style.cssText = this.properties[7];
		if( this.properties[8] )
		{
			var wwCSSheet		= document.createElement( 'style' );
			wwCSSheet.innerHTML	= this.properties[8];
			document.body.appendChild( wwCSSheet );
		}

		this.jqelem.appendTo(this.runtime.canvasdiv ? this.runtime.canvasdiv : "body");
		
		this.elem.title = this.properties[2];
		this.elem.disabled = (this.properties[4] === 0);
		
		this.autoFontSize = (this.properties[5] !== 0);
		this.element_hidden = false;
		
		if (this.properties[3] === 0)
		{
			this.jqelem.hide();
			this.visible = false;
			this.element_hidden = true;
		}
		
		this.elem.onclick = (function (self) {
			return function(e) {
				e.stopPropagation();
				
				self.runtime.isInUserInputEvent = true;
				self.runtime.trigger(cr.plugins_.HtmEl.prototype.cnds.OnClicked, self);
				self.runtime.isInUserInputEvent = false;
			};
		})(this);
		
		// Prevent touches reaching the canvas
		this.elem.addEventListener("touchstart", function (e) {
			e.stopPropagation();
		}, false);
		
		this.elem.addEventListener("touchmove", function (e) {
			e.stopPropagation();
		}, false);
		
		this.elem.addEventListener("touchend", function (e) {
			e.stopPropagation();
		}, false);
		
		// Prevent clicks being blocked
		this.jqelem.mousedown(function (e) {
			e.stopPropagation();
		});
		
		this.jqelem.mouseup(function (e) {
			e.stopPropagation();
		});
		
		// Prevent key presses being blocked by the Keyboard object
		this.jqelem.keydown(function (e) {
			e.stopPropagation();
		});
		
		this.jqelem.keyup(function (e) {
			e.stopPropagation();
		});
		
		this.lastLeft = 0;
		this.lastTop = 0;
		this.lastRight = 0;
		this.lastBottom = 0;
		this.lastWinWidth = 0;
		this.lastWinHeight = 0;
			
		this.updatePosition(true);
		
		this.runtime.tickMe(this);
	};
	
	instanceProto.saveToJSON = function ()
	{
		var o = {
			"tooltip": this.elem.title,
			"disabled": !!this.elem.disabled
		};
		o["innerHTML"] = this.elem.innerHTML;
		return o;
	};
	
	instanceProto.loadFromJSON = function (o)
	{
		this.elem.title = o["tooltip"];
		this.elem.disabled = o["disabled"];
		this.elem.innerHTML = o["innerHTML"];
	};
	
	instanceProto.onDestroy = function ()
	{
		if (this.runtime.isDomFree)
			return;
		
		this.jqelem.remove();
		this.elem	= null;
		this.jqelem		= null;
	};
	
	instanceProto.tick = function ()
	{
		this.updatePosition();
	};
	
	
	instanceProto.updatePosition = function (first)
	{
		if (this.runtime.isDomFree)
			return;
		
		var left = this.layer.layerToCanvas(this.x, this.y, true);
		var top = this.layer.layerToCanvas(this.x, this.y, false);
		var right = this.layer.layerToCanvas(this.x + this.width, this.y + this.height, true);
		var bottom = this.layer.layerToCanvas(this.x + this.width, this.y + this.height, false);
		
		// Is entirely offscreen or invisible: hide
		if (!this.visible || !this.layer.visible || right <= 0 || bottom <= 0 || left >= this.runtime.width || top >= this.runtime.height)
		{
			if (!this.element_hidden)
				this.jqelem.hide();
			
			this.element_hidden = true;
			return;
		}
		
		// Truncate to canvas size
		if (left < 1)
			left = 1;
		if (top < 1)
			top = 1;
		if (right >= this.runtime.width)
			right = this.runtime.width - 1;
		if (bottom >= this.runtime.height)
			bottom = this.runtime.height - 1;
		
		var curWinWidth = window.innerWidth;
		var curWinHeight = window.innerHeight;
			
		// Avoid redundant updates
		if (!first && this.lastLeft === left && this.lastTop === top && this.lastRight === right && this.lastBottom === bottom && this.lastWinWidth === curWinWidth && this.lastWinHeight === curWinHeight)
		{
			if (this.element_hidden)
			{
				this.jqelem.show();
				this.element_hidden = false;
			}
			
			return;
		}
			
		this.lastLeft = left;
		this.lastTop = top;
		this.lastRight = right;
		this.lastBottom = bottom;
		this.lastWinWidth = curWinWidth;
		this.lastWinHeight = curWinHeight;
		
		if (this.element_hidden)
		{
			this.jqelem.show();
			this.element_hidden = false;
		}
		
		var offx = Math.round(left) + jQuery(this.runtime.canvas).offset().left;
		var offy = Math.round(top) + jQuery(this.runtime.canvas).offset().top;
		this.jqelem.css("position", "absolute");
		this.jqelem.offset({left: offx, top: offy});
		this.jqelem.width(Math.round(right - left));
		this.jqelem.height(Math.round(bottom - top));
		
		if (this.autoFontSize)
			this.jqelem.css("font-size", ((this.layer.getScale(true) / this.runtime.devicePixelRatio) - 0.2) + "em");
	};
	
	// only called if a layout object
	instanceProto.draw = function(ctx)
	{
	};
	
	instanceProto.drawGL = function(glw)
	{
	};
	
	/**BEGIN-PREVIEWONLY**/
	instanceProto.getDebuggerValues = function (propsections)
	{
		propsections.push({
			"title": "HtmEl",
			"properties": [
				{"name": "innerHTML", "value": this.isCheckbox ? this.labelText.nodeValue : this.elem.value}
			]
		});
	};
	
	///	TODM not sure why and what is this:
	instanceProto.onDebugValueEdited = function (header, name, value)
	{
		if (name === "innerHTML")
		{
			this.elem.innerHTML = value;
		}
	};
	/**END-PREVIEWONLY**/







	//    //\\    Conditions    //////////////////////////////////
	function Cnds() {};
	
	Cnds.prototype.OnClicked = function ()
	{
		return true;
	};
	pluginProto.cnds = new Cnds();
	//    \\//    Conditions    //////////////////////////////////
	





	//    //\\    Actions    //////////////////////////////////
	function Acts() {};

	Acts.prototype.SetInnerHTML = function (text)
	{
		if (this.runtime.isDomFree)
			return;
		
		this.elem.innerHTML = text;
	};
	
	Acts.prototype.SetTooltip = function (text)
	{
		if (this.runtime.isDomFree)
			return;
		
		this.elem.title = text;
	};
	
	Acts.prototype.SetVisible = function (vis)
	{
		if (this.runtime.isDomFree)
			return;
		
		this.visible = (vis !== 0);
	};
	
	Acts.prototype.SetEnabled = function (en)
	{
		if (this.runtime.isDomFree)
			return;
		
		this.elem.disabled = (en === 0);
	};
	
	Acts.prototype.SetFocus = function ()
	{
		if (this.runtime.isDomFree)
			return;
		
		this.elem.focus();
	};
	
	Acts.prototype.SetBlur = function ()
	{
		if (this.runtime.isDomFree)
			return;
		
		this.elem.blur();
	};
	
	Acts.prototype.SetCSSStyle = function ( pp, vv)
	{
		if (this.runtime.isDomFree)
			return;
			
		this.jqelem[ 'css' ]( pp, vv);
	};
	
	Acts.prototype.SetInlineCssText = function ( cssText )
	{
		if (this.runtime.isDomFree)
			return;
			
		this.elem.style.cssText = cssText;
	};










	Acts.prototype.elMethod = function ( methodPath, args, destinationPath )
	{
		callFunction( this, this.elem, methodPath, args, destinationPath );
	};

	Acts.prototype.elSet = function ( propPath, string )
	{
		if (this.runtime.isDomFree) return;
		set( this.elem, propPath, string );
	};

	///	calls like: jQuery( this.elem ).command
	Acts.prototype.jqel = function ( methodPath, args )
	{
		callFunction( this, this.jqelem, methodPath, args );
	};

	///	calls and assigns returned object;
	Acts.prototype.jq = function ( methodPath, args, destinationPath )
	{
		callFunction( this, jQuery, methodPath, args, destinationPath );
	};

	Acts.prototype.coll = function ( funPath, args, destinationPath )
	{
		callFunction( this, window, funPath, args, destinationPath );
	};

	Acts.prototype.assByPath = function ( destinationPath, departurePath )
	{
		var value = get( window, departurePath );
		set( window, destinationPath, value );
		this.ret = value;
		//	c onsole.log( 'this.jsob = ', this.jsob );
	}

	Acts.prototype.assPrimitive = function ( destinationPath, value )
	{
		set( window, destinationPath, value );
		this.ret = value;
		//	c onsole.log( 'this.jsob = ', this.jsob );
	}

	Acts.prototype.ret2path = function ( destinationPath, departurePath )
	{
		var value = get( this.ret, departurePath );
		set( window, destinationPath, value );
		//	c onsole.log( 'this.jsob = ', this.jsob );
	}

	//	converts json to object and assings this object to node window.path
	Acts.prototype.json2path = function ( json, path )
	{
		try {
			var value = JSON.parse( json );
			//.	"stealthy leaves"
			//	c onsole.log( 'object from json=', value );
			if( typeof value !== 'object' || value === null ) return;
			set( window, path, value );
		}
		catch(e)
		{
			//	c onsole.log( "error" + e );
			return;
		}
	};


	///	injects global css ( cross domain ? )
	Acts.prototype.appendCSS = function ( path )
	{
		var sc		= document.createElement( 'link' );
		sc.type		= 'text/css';
		sc.rel		= 'stylesheet';
		sc.href		= path;

		//. TODM why not simply as:
		document.head.appendChild( sc );

		//var sib		= document.getElementsByTagName( 'link' ) && document.getElementsByTagName( 'link' )[ 0 ];
		//assumed: getElementsByTagName is permitted to be an empty collection;
		//document.body.appendChild( sc );
		//sib.parentNode.insertBefore( sc, sib );		
	};


	///	injects global css from text
	Acts.prototype.appendCSSFromText = function ( css_text )
	{
			var CSSheet			= document.createElement( 'style' );
			CSSheet.innerHTML	= css_text;
			document.body.appendChild( CSSheet );
	};

	pluginProto.acts = new Acts();
	//    \\//    Actions    //////////////////////////////////





	
	//    //\\    Expressions    //////////////////////////////////
	function Exps() {};
	Exps.prototype.getInlineCssText = function ()
	{
		ret.set_string( this.elem.style.cssText );
	};




	///	gets from path relative to window
	Exps.prototype.path = function ( ret, path )
	{
		var val = get( window, path );
		var type = typeof val;
		if( type === 'string' )
		{
			ret.set_string( val );
		} else if( type === 'number' ) {
			ret.set_any( val );
		} else {
			///	falls back to 0 or 1 depnding on !!val because cannot make c2 to return objects;
			ret.set_any( val ? "1" : "" );
		}
	};

	///	gets this.path, stashes in this.ret, and returns as primitive if any;
	Exps.prototype.get = function ( ret, path )
	{
		var val = get( this.elem, path );
		this.ret = val;

		var type = typeof val;
		if( type === 'string' )
		{
			ret.set_string( val );
		} else if( type === 'number' ) {
			ret.set_any( val );
		} else {
			///	falls back to 0 or 1 depnding on !!val because cannot make c2 to return objects;
			ret.set_any( val ? "1" : "" );
		}
	};


	Exps.prototype.coll = function ( ret, funPath, args )
	{
		callFunction( this, window, funPath, args );
		returnRet( this, ret );
	};


	///	runs method, stashed result in this.ret, and returns result.path;
	Exps.prototype.method = function ( ret, methodPath, args )
	{
		callFunction( this, this.elem, methodPath, args );
		returnRet( this, ret );
	};


	///	returns this.ret.path where this.ret is a result of last ran method;
	Exps.prototype.ret = function ( ret, path )
	{
		returnRet( this, ret, path );
	};


	Exps.prototype.jqel = function ( ret, methodPath, args )
	{
		callFunction( this, this.jqelem, methodPath, args );
		returnRet( this, ret );
	};

	Exps.prototype.jq = function ( ret, methodPath, args )
	{
		callFunction( this, jQuery, methodPath, args );
		returnRet( this, ret );
	};

	
	Exps.prototype.path2json = function ( ret, path )
	{
		var ww = get( window, path );
		//	c onsole.log( path, ww );
		if( typeof ww !== 'object' || ww === null ) ww = {};
		var ww = JSON.stringify( ww, null, "\t" );
		ret.set_string( ww );
	};




	pluginProto.exps = new Exps();
	//    \\//    Expressions    //////////////////////////////////








	//    //\\    Helpers    //////////////////////////////////
	var returnRet = function ( thiss, ret, departurePath )
	{
		if (thiss.runtime.isDomFree)
		{
			ret.set_string( '' );
			return;
		}
		var val = get( thiss.ret, departurePath );
		var type = typeof val;
		if( type === 'string' )
		{
			ret.set_string( val );
		} else if( type === 'number' ) {
			ret.set_any( val );
		} else {
			///	falls back to 0 or 1 depnding on !!val because cannot make c2 to return objects;
			ret.set_any( val ? "1" : "" );
		}
	};

	///	calls/sets method of element or method of jQ-wrapper of the element or jQuery itself
	var callFunction = function ( thiss, methodParentObj, methodPath, args, destinationPath )
	{
		if (thiss.runtime.isDomFree) return !'success';
		var res			= get( methodParentObj, methodPath, 'WithParent' ); 
		var fun			= res[ 'obj' ];
		var parent		= res[ 'parent' ];
		var operands	= args.split( RIN_SEPARATOR );
		if( typeof fun	=== 'function' && parent )
		{
			thiss.ret = fun.apply( parent, operands );
		}
		//	c onsole.log ( ' methodParentObj=' + methodParentObj + ' fun=', fun );
		
		if( destinationPath )
		{
			set( null, destinationPath, thiss.ret );
		}
		return 'success';
	};

	////	TODM delete; possibly, one wants to set a property, do this
	//		thiss.ret = elem[ methodName ] = operands[ 0 ] || '';

	//    \\//    Helpers    //////////////////////////////////







	//\\//	Own microlibrary


	/// Acts:		extends object's property-tree and assigns value to a leaf-property
	//
	//	Usage:		propertify( conf, 'startup.items', 4 ) assigns
	//				conf.startup.items = value "starup" initially does not exist.
	//
	//	Input:		obj - top object, as in above obj = conf;
	var set = function ( obj, path, value )
	{
		var tokens	= path.split( '.' );
		var len		= tokens.length;
		var len1	= len - 1;
		if( len1 < 0 ) return;
		//.	be careful with this:
		if( typeof obj !== 'object' || obj === null ) obj = window;
		var prop	= tokens[ 0 ];

		for( var ii = 0; ii < len1; ii++ )
		{
			//// appends objects if missed
			if( !obj[ prop ] || typeof obj[ prop ] !== 'object' ) obj[ prop ] = {};
			obj = obj[ prop ];
			var prop = tokens[ ii + 1 ];
		}
		obj[ prop ] = value;
	};



	///	returns value from obj[ path ] even path is like path = level.child3.node2. ...
	var get = function ( obj, path, withParent )
	{
		var ret		= withParent ? { "parent" : null, "obj" : obj } : obj;
		if( !path )	return ret;
		var tokens	= path.split( '.' );
		var len		= tokens.length;
		var len1	= len - 1;
		if( len1 < 0 ) return ret;
		var prop	= tokens[ 0 ];

		if( typeof obj !== 'object' || obj === null ) obj = window;

		for( var ii = 0; ii < len1; ii++ )
		{
			var parent = obj;
			obj = obj[ prop ];
			if( typeof obj !== 'object' || obj === null )
			{
				return withParent ? { "parent" : parent, "obj" : obj } : obj;
			}
			var prop = tokens[ ii + 1 ];
		}
		// if( prop === 'length' ) c onsole.log( obj, ' length=' + obj[ prop ] );
		return withParent ? { "parent" : obj, "obj" : obj[ prop ] } : obj[ prop ]; 
	};


}());
